/********************************************************************* 
  
   fasttrak.c -- PROMISE FastTrak Controllers device driver for Linux.

   Copyright (C) 2002-2005 PROMISE Technology, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   Bugs/Comments/Suggestions should be mailed to:                            
   support@promise.com.tw

   For more information, goto:
   http://www.promise.com

*********************************************************************/

#include <stdarg.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/byteorder.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/blk.h>
#include <linux/tqueue.h>
#include <linux/types.h>
#include <linux/hdreg.h>
#include <linux/wrapper.h>
#include <linux/module.h>
#include <linux/reboot.h>
#include <linux/fs.h>
#include <linux/poll.h>
 
#include "/usr/src/linux/drivers/scsi/sd.h"
#include "/usr/src/linux/drivers/scsi/scsi.h"
#include "/usr/src/linux/drivers/scsi/hosts.h"

#include "global.h"
#include "fasttrak.h"

ScsiEngReq_p	CurScsiReq;
unchar		*FT_PHY[FT_Max_Ada], *FT_VIR[FT_Max_Ada];
int 		MajorNumber;	/* for ft3xxx device file */
extern int nForce_present;		/* Fix nForce chipset */

/* for FT_PHY/FT_VIR to __get_free_pages */
#define	PowerOfPHY 	7	/* 4K*128 = 512KB */
#define	PowerOfVIR	4	/* 4K*16 =  64KB */

MODULE_AUTHOR ("Promise Inc.");
#ifdef __SMP__
MODULE_DESCRIPTION ("PROMISE FastTrak TX4000/376/378/S150 TX Series Linux Driver (SMP)");
#else
MODULE_DESCRIPTION ("PROMISE FastTrak TX4000/376/378/S150 TX Series Linux Driver");
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10)
MODULE_LICENSE("Proprietary");
//MODULE_LICENSE("GPL");
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
   static DECLARE_WAIT_QUEUE_HEAD(Wait24);
   static DECLARE_WAIT_QUEUE_HEAD(Wait24SmartCmd);
#else   
   static struct wait_queue *Wait22=NULL;
   static struct wait_queue *Wait22SmartCmd=NULL;
#endif   
   
struct proc_dir_entry proc_scsi_ft3xx = {		
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
   	0,				// PROC_SCSI_FASTTRAK,
#else
   	PROC_SCSI_NOT_PRESENT,		// PROC_SCSI_FASTTRAK,
#endif
   	5,				// length of "ft3xx" 	
   	"ft3xx",			// name of directory entry
   	S_IFDIR | S_IRUGO | S_IXUGO,	
   	2,				
};

/*********************************************************************
 * Function:
 *   int ft3xx_proc_info(char *, char **, off_t, int, int, int)
 *
 * Description:
 *   in user space , user can view "/proc/scsi/ft3xx/x" 
 *   to run ft3xx_proc_info() in kernel space.
 * PS: 
 *   the rebuilding information reported under the condition of
 *   one MIRROR array , not support two MIRROR array concurrently.
*********************************************************************/
int ft3xx_proc_info(char *buffer, char **start, off_t offset,
			int length, int hostno, int func)
{
	ft_proc_t info;

	info.buffer = buffer;
	info.length = length;
	info.offset = offset;
	info.position = 0;
	if (start) {
		*start = buffer;
	}

	GetProcInfo(&info);

	if (info.position > info.offset) {
		return (info.position - info.offset);
	} else {
		return 0;
	}

}

/*********************************************************************
 * Function:
 * 	int ft3xx_ioctl(Scsi_Device *device,int cmd,void *arg)
 *
 * Description:
 *********************************************************************/
static int ft3xx_ioctl(Scsi_Device *device,int cmd,void *arg)
{
	switch( cmd ) {
		case 0x024B:	// Red Hat Ready Test Suite: rhr-fdisk
		case 0x5310:	// cdrom_multisession(SuSE Linux)
	   		break;
		default:
			printk(KERN_WARNING "ft3xx: Unsupported IOCTL CMD");
			printk(KERN_WARNING "(%#x)\n",cmd);
   	}
	return(0);
}

/***********************************************************
 * Description:
 *	FastTrak(ft3xx) Character Device 
 *	File(ft3xx) operations for FastTrak RAID-API
 * ---------------------------------------------------------
 * int ftdev_open(struct inode *inodep, struct file *filep) 
 * ssize_t ftdev_read (struct file *filp, char *buf, 
 *		size_t count,loff_t *f_pos)
 * int ftdev_release(struct inode *inodep, struct file *filep) 
 * int ftdev_ioctl(struct inode *inodep, struct file *filep, 
 *		unsigned int req, unsigned long arg) 
 * int ftdev_fasync(int fd, struct file *filp, int mode)
***********************************************************/
int ftdev_open(struct inode *inodep, struct file *filep) 
{
	if( (MajorNumber <= 0) || (MajorNumber != MAJOR(inodep->i_rdev)) )
		return(-ENODEV);

	if (!filep->private_data) {
		filep->private_data = &ftdev_fd;
	}

	MOD_INC_USE_COUNT;
	return(0);
}

ssize_t ftdev_read (struct file *filp, char *buf, size_t count,
		                loff_t *f_pos)
{
	count = 20;
	copy_to_user(buf, "ft3xx: ftdev_read()\n", count);
	return count;
}

int ftdev_release(struct inode *inodep, struct file *filep) 
{
	int ftdev_fasync(int fd, struct file *filp, int mode);

	ftdev_fasync(-1, filep, 0);
	MOD_DEC_USE_COUNT;
	return(0);
}

extern int do_ftdev_ioctl(unsigned int req, unsigned long arg) ;
int ftdev_ioctl(struct inode *inodep, struct file *filep, 
		unsigned int req, unsigned long arg) 
{
	return( do_ftdev_ioctl(req , arg) ) ;
}

int ftdev_fasync(int fd, struct file *filp, int mode)
{
	ftdev_fd_t *dev = filp->private_data;

	return fasync_helper(fd, filp, mode, &dev->async_queue);
}

#if 0
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
struct block_device_operations ftdev_fops = {
	open:		ftdev_open, 
	release:	ftdev_release, 
	ioctl:		ftdev_ioctl,
};
#else
struct file_operations ftdev_fops = {
	NULL, NULL, NULL, NULL, NULL, ftdev_ioctl, NULL,
	ftdev_open, NULL, ftdev_release, NULL, NULL, NULL, NULL};
#endif
#endif

/* ft3xx: Character Device File Operations */
struct file_operations ftdev_fops = {
	open:		ftdev_open,
	read:		ftdev_read,
	release:	ftdev_release,
	ioctl:		ftdev_ioctl,
	fasync:		ftdev_fasync,
};

/*********************************************************************
 * Function:
 *   static int ft_halt(struct notifier_block *nb, ulong event, void *buf)
 *
 * Description:
 *   Notifier block to get a notify on system shutdown/halt/reboot 
 ********************************************************************/
static int ft_halt(struct notifier_block *nb, ulong event, void *buf);
static struct notifier_block ft_notifier = {
	ft_halt, NULL, 0
};

/*********************************************************************
 * Function:
 *   static int ft_halt(struct notifier_block *nb, ulong event, void *buf)
 *
 * Description:
 *   The function will been called when "ft3xx_release" executed and
 *   system reboot/halt(register_reboot_notifier(&ft_notifier)).
*********************************************************************/
extern int ft_shutdown_controllers(void);
static int ft_halt(struct notifier_block *nb, ulong event, void *buf)
{
	/* shutdown PROMISE controller(s) and free resource */
	ft_shutdown_controllers();
	unregister_reboot_notifier(&ft_notifier);
	return NOTIFY_OK;
} 

/*********************************************************************
 * Function:
 *   int SleepWakeUp(int fun) ;
 *
 * Description:
 ********************************************************************/
int SleepWakeUp(int id, int fun) 
{
	switch( fun ) {
		case 0 :	// Sleep
	   		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
			if( id == 1 ) {
				interruptible_sleep_on(&Wait24);
			} else if( id == 2 ) {
				interruptible_sleep_on(&Wait24SmartCmd);
			}
	   		#else
			if( id == 1 ) {
				module_interruptible_sleep_on(&Wait22);
			} else if( id == 2 ) {
				module_interruptible_sleep_on(&Wait22SmartCmd);
			}
	   		#endif
	   		break;
		case 1 :	// Wake Up		
	   		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
			if( id == 1 ) {
				wake_up_interruptible(&Wait24);
			} else if( id == 2 ) {
				wake_up_interruptible(&Wait24SmartCmd);
			}
	   		#else
			if( id == 1 ) {
				module_wake_up(&Wait22);
			} else if( id == 2 ) {
				module_wake_up(&Wait22SmartCmd);
			}
	   		#endif
	   		break;
		default :
	   		printk(KERN_ERR "ft3xx: SleepWakeUP(%d) Error\n",fun);
	}
	return(0);
}

/*********************************************************************
 * Function:
 *   void FreeResource(char force, unchar *str) ;
 *
 * Description:
 *   Free Memory Resource and print message if need .
 ********************************************************************/
void FreeResource(char force, unchar *str) 
{
	int	i;
	for( i=0 ; i < FT_Max_Ada ; i++ ) {
		if( ft_adapter[i].present ) {
			free_pages((ulong)FT_PHY[i],PowerOfPHY);
			free_pages((ulong)FT_VIR[i],PowerOfVIR);
		}
	}
	if( force ) {
		printk(KERN_DEBUG "%s",str);
	} else {
		DEB(printk(KERN_DEBUG "%s",str););
	}
	return;
}

/*********************************************************************
 * Function:
 *   void ft_handler(int irq, void *dev_id, struct pt_regs *regs) 
 *
 * Description:
 *   ISR of ft3xx. As a matter of fact, if multiple FastTraks don't 
 *   share IRQ, we can get adapter number from argument "dev_id".
 ********************************************************************/
void ft_handler(int irq, void *dev_id, struct pt_regs *regs) 
{
	int	ada ;
	ulong	smpflag;
#if 0 
	printk("Adapter IRQ[%d/%d]: %d\n",ft_adapter[0].ft_pci_dev->irq,ft_adapter[1].ft_pci_dev->irq,irq);  
#endif
	spin_lock_irqsave(&io_request_lock,smpflag);
	for( ada=0 ; ada < FT_Max_Ada ; ada++ ) {
   		if( ft_adapter[ada].present ) { 
   	   		if( irq == ft_adapter[ada].ft_pci_dev->irq ) {
				CAM_ISR( 1 << ada ) ;
	   		}
		}
   	}
	spin_unlock_irqrestore(&io_request_lock,smpflag);
	return;
}

/*********************************************************************
 * Function:
 *   unchar init_ioports( unchar ada )
 *
 * Description:
 *   ft3xx ioports initialization
 ********************************************************************/
int init_ioports( unchar ada )
{
	unchar 		uc;
	ft_adapter_t 	*adapter = &ft_adapter[ada] ;
	struct pci_dev	*ft_pci_dev = adapter->ft_pci_dev;
	ushort		command ;
	uint 		base[6];

	for( uc = 0 ; uc < 6 ; uc++ ) {
     		uint	range = 0 ;
     		uint	base0 = PCI_BASE_ADDRESS_0 ;
     		pci_read_config_dword (ft_pci_dev,base0+4*uc,&base[uc]);
     		pci_write_config_dword(ft_pci_dev,base0+4*uc,0xffffffff);
		pci_read_config_dword (ft_pci_dev,base0+4*uc,&range);
		range = (uint) ~range + 1;
		/* fix base0/base1/base2 range */
		adapter->range[uc] = range + ((uc < 3) ? 1 : 0) ;   
		pci_write_config_dword(ft_pci_dev,base0+4*uc,base[uc]);
		adapter->base[uc] = base[uc] & PCI_BASE_ADDRESS_IO_MASK ;
   	}
	
   	{
	/* for ftdev IOCTL Interface/PAM */
		extern uint	gBaseAddress3 ;
		extern uint	gBaseAddress4 ;
		gBaseAddress3 = adapter->base[3] ;
		gBaseAddress4 = adapter->base[4] ;
	}	   
	adapter->base[3] = (uint) ioremap(adapter->base[3],adapter->range[3]) ;
	adapter->base[4] = (uint) ioremap(adapter->base[4],adapter->range[4]) ;
  
	pci_read_config_word(ft_pci_dev,PCI_COMMAND,&command);
	command |= PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY;
	pci_write_config_word(ft_pci_dev,PCI_COMMAND,command);
	pci_read_config_dword(ft_pci_dev,0x30,&adapter->rom_base);

	/* Request I/O ports */
	for( uc = 0 ; uc < 3 ; uc++ ) {
   		if( check_region(adapter->base[uc],adapter->range[uc]) ) {
			return( 0x10 | uc ) ;
   		}
	}
	request_region(adapter->base[0],adapter->range[0],"ft3xx0");
	request_region(adapter->base[1],adapter->range[1],"ft3xx1");
	request_region(adapter->base[2],adapter->range[2],"ft3xx2");
   
	/* Request IRQ & Install an interrupt handler */
	/* Check IRQ number first if has been used by FastTrak */
	for( uc=0 ; uc < ada ; uc++ ) {
		if( ft_adapter[uc].ft_pci_dev->irq == ft_pci_dev->irq ) {
			return(2);
		}
	}
	if( request_irq(ft_pci_dev->irq,ft_handler,
		SA_INTERRUPT | SA_SHIRQ,"ft3xx",&ft_adapter[ada])) {
   		printk(KERN_ERR "ft3xx: adapter(%d) request irq(%d) "
			"error.\n",ada+1,ft_pci_dev->irq);
   		return(2);
	}
   	return(0);
}

/*********************************************************************
 * Function:
 *   FindAdaptersNumber()
 *
 * Description:	for Linux Kernel 2.2.x & 2.4.x
 *   1. find controller
 *   2. find ft3xx adapter ( ClassCode=0x0104 ) 
*********************************************************************/
int FindAdaptersNumber( void ) 
{
	unchar		class ;
	unchar		adapters ;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	struct pci_dev	*ft_pci_dev = NULL ;
#else
	struct pci_dev	*ft_pci_dev = pci_devices ;
#endif
	unsigned short	nVidiaVID = 0x10de ;
	unsigned int	nVidiaCLASS = 0x060000 ;
	class = adapters = 0 ;	
	while( FT_DEVID[class] ) {
   		while( (ft_pci_dev = pci_find_device(PCI_VENDOR_ID_PROMISE,FT_DEVID[class],ft_pci_dev)) ) {
			ft_adapter_t	*pada = &ft_adapter[adapters];
	   		if( (ft_pci_dev->class) >> 8 != PCI_CLASS_STORAGE_RAID) 
				continue;
	   		if( adapters >= FT_Max_Ada ) {
		   		printk(KERN_ERR 
					"ft3xx: The driver only supports "
					"one/two ft3xx adapter(s).\n"
					"Others do not be supported:\n");
   		   		return(adapters);
			}
		   	/* PCI_CACHE_LINE_SIZE: FastTrak=0x90 SATA=0x91 */
			if( ft_pci_dev->device == DID378 ) {
		   		unchar	diff ;
     		   		pci_read_config_byte(ft_pci_dev,PCI_CACHE_LINE_SIZE,&diff);
		   		if( diff != 0x90) {	
					printk(KERN_ERR 
						"ft3xx: The driver doesn't "
						"support SATA 378\n");
					continue;
		   		}
			}
   			pada->present 	= 1;
			pada->ft_pci_dev= ft_pci_dev;
			pada->deviceid	= ft_pci_dev->device;
			pada->irq	= ft_pci_dev->irq;
	   		adapters ++;
   		}
		class++;
	}	// End of while( FT_DEVID[class] )
	/* >> find nForce chipset for fix */
	if ( adapters ) {
		while ( (ft_pci_dev = pci_find_device(nVidiaVID, PCI_ANY_ID, ft_pci_dev) ) ) {
			if ( ft_pci_dev->class == nVidiaCLASS )
				nForce_present = 1;
		}
	}
	/* Add by Hank 2003.08.19 << */
	return(adapters);
}

/*********************************************************************
 * Function:
 *   FindAdapters()
 *
 * Description:
 *   1. find controller
 *   2. find ft3xx adapter ( ClassCode=0x0104 ) 
 *   3. request ioports and irq , and then fill out structure
*********************************************************************/
int FindAdapters( void ) 
{
	unchar		uc ;
	unchar		adapters ;

	if ( ! pci_present() ) {	// return non-zero on detection
		DEB(printk(KERN_WARNING "ft3xx: PCI card isn't present.\n"));
		return 0;
	}
	if( !(adapters = FindAdaptersNumber()) )
		return(0);
   
	for( uc=0 ; uc < FT_Max_Ada ; uc++ ) {
		if( ft_adapter[uc].present ) { 
   	   		if( init_ioports(uc) ) {
				printk(KERN_ERR "ft3xx: Initial "
					"Adapter%d IO-Port error.\n",uc+1);
				return(0);
	   		}
		}
	}
	return(adapters);
}

/*********************************************************************
 * Function:
 *   do_ft3xx_detect
 *
 * Description:
 *   Try to detect , initialize and register an ft3xx adapter.
 *   Return non-zero number on detection.
 *   The driver only support one ft3xx adapter now.
*********************************************************************/
int do_ft3xx_detect( struct SHT *template ) 
{
	unchar			adapters ;
	int			i ;
	struct Scsi_Host	*shpnt ;
	ft_adapter_t		*ft_ada;
	ulong			PSize,VSize;
   
	if( GetEngNeededMemSize(&PSize,&VSize) )
		return(0);

	/*	
	 *	Initial Global Data Structure : 
	 *		ft_adapter[FT_Max_Ada]
	 *		ft_channel[FT_Max_Ch]
	 *		ft_drive[FT_Max_Dev]
	 */
	if( InitGDS() ) {
   		printk(KERN_ERR "ft3xx: Initial "
			"Global Data Structure fail.\n");
   		return(0);
	}

	/*	
	 *	Find Adapter :
	 *		. find controller
	 *		. find ft3xx adapter ( ClassCode=0x0104 ) 
	 *		. request ioports and irq, and then fill out structure
	 */
	if( !(adapters = FindAdapters()) )
		return(0);

	/* 
	 *    Register Timer for ft3xx
	 */ 	
	RegisterTimer();

 	/*	
	 * 	Initialize CAM & TaskQue
	 */
	if( OpenCAM() ) {
   		printk(KERN_ERR "ft3xx: OpenCAM() fail\n");
   		return(0);
	}

	/*
	 *	Initialize all exist PROMISE Adapters
	 *		- Initialize RAID ENGINE 
	 * 		- Register SCSI hosts 
	 */
	for( i=0 ; i < adapters ; i++ ) {
		/*
		 *	Initialize RAID ENGINE :
		 *		1. allocate memory for RAID ENGINE arguments
		 *		2. start initialize RAID ENGINE :  bInitRAID
		 */
		/* physical:128KB virtual:64KB */
		FT_PHY[i] = (unchar *) __get_free_pages(GFP_ATOMIC,PowerOfPHY);
		FT_VIR[i] = (unchar *) __get_free_pages(GFP_ATOMIC,PowerOfVIR);
		if( (!FT_PHY[i]) || (!FT_VIR[i]) ) {
			printk(KERN_ERR "ft3xx: Insufficient System Memory\n");
			printk(KERN_ERR "ft3xx: ft3xx Driver Module removed\n");
			return(0);
		}
		memset(FT_PHY[i], 0, PAGE_SIZE << PowerOfPHY);
		memset(FT_VIR[i], 0, PAGE_SIZE << PowerOfVIR);

		bInitRAIDTimer(i) ;		
		bInitRAID(bPOWER_UP,virt_to_bus(FT_PHY[i]),FT_VIR[i],PSize,VSize,i);
		SleepWakeUp(1,0) ;	// Sleep

		/*	
		 * 	Register Array :
		 *		register arrays to OS not drives to OS.
		 *		One adapter supports four arrays.
		 *		Two adapters support eight arrays.
		 *		However, array components can't been setup
		 *		on seprated adapters.
		 */
		shpnt = scsi_register(template,sizeof(ft_adapter_t));
		shpnt -> max_channel = 0 ;	/* channel: primary */
		shpnt -> max_id = 4 ;		/* dev : array[0-3] */
		shpnt -> max_lun = 1 ;		/* one device       */
		if( !shpnt->hostdata ) {
			printk(KERN_ERR "ft3xx: Bad scsi hostdata(%d).\n",i);
			/* scsi_unregister(shpnt); release ..... */
			return(0);
		}
	   	ft_ada = (ft_adapter_t *) shpnt->hostdata ;
   		memcpy(ft_ada,&ft_adapter[i],sizeof(ft_adapter_t));
  	}

	/*
	 *	Fill out data structure after RAID initialize
	 *		1. define ft_drive[dev].arr  
	 *		2. define ft_drive[dev].strm 
	 */
	if( DefDriveStruct() ) {
	   	FreeResource(1,"ft3xx: DefDriveStruct() Error.\n");
   		return(0);
	}
   
	CurScsiReq = kmalloc(sizeof(*CurScsiReq),GFP_ATOMIC);
	InitQueueBuffer();

	{ 
		extern unchar GetRebuildRate(void);
		extern ushort GetSyncRate(void);
		extern unchar SetRebuildRate(unchar value);
		extern unchar SetSyncRate(unchar value);
		SetRebuildRate(4) ; 
		SetSyncRate(4);
	}

	return(adapters);
} 

/*********************************************************************
 * Function:
 *   ft3xx_detect
 *
 * Description:
 *   Try to detect , initialize and register an ft3xx adapter.
 *   Return non-zero number on detection.
*********************************************************************/
int ft3xx_detect( struct SHT *template )
{
	unchar		adapters ;

	printk(KERN_WARNING "PROMISE FastTrak TX4000/376/378/S150 TX Series "
		"Linux Driver Version " VERSION " " VERDATE"\n");
	DebScsiFun(printk(KERN_DEBUG "DebScsiFun: ft3xx_detect(). \n"););
	template -> proc_dir   = &proc_scsi_ft3xx;
	template -> proc_info  = ft3xx_proc_info;     
	adapters = do_ft3xx_detect( template );

	if( adapters ) {
		register_reboot_notifier(&ft_notifier);
	
		/* register ft3xx device as Character Device */
#if 0
		if( (MajorNumber = register_blkdev( 0, "ft3xx", 
			(struct block_device_operations *)&ftdev_fops)) <= 0){
#endif
		if ( (MajorNumber = register_chrdev( 0, "ft3xx",
			(struct file_operations *)&ftdev_fops)) <= 0) {
			printk(KERN_ERR "ft3xx: register "
				"ft3xx device error(%d)",MajorNumber);
		}
	} else {
		printk(KERN_WARNING "ft3xx: No valid controller detected\n");
	}	   

	return(adapters);
}

/*********************************************************************
 * Function:
 *   ft3xx_release
 *
 * Description:
 *   The function will been called when "scsi_unregister_module" 
 *   executed on Linux System "rmmod". It needs only to release
 *   "Scsi_Host" resource, not "Scsi_Host_Template". However, ft3xx
 *   release all resource at once now.
*********************************************************************/
int ft3xx_release(struct Scsi_Host *host)
{
	static int release_counter = 0 ;
	if( release_counter++ )
		return(0);
#if 0
   	unregister_blkdev( MajorNumber , "ft3xx" );
#endif
   	unregister_chrdev( MajorNumber , "ft3xx" );
	ft_halt(0,0,0);
	FreeResource(0,"");
	scsi_unregister(host);
	return(0);
}

/***********************************************************
 * Function:
 *   ft3xx_queuecommand
 *
 * Description:
 *   cmd - SCSI command.
 *   fn  - function called on completion.
***********************************************************/
extern int do_queuecommand(ScsiEngReq_p p);	
int ft3xx_queuecommand(Scsi_Cmnd *cmd, void (*fn)(Scsi_Cmnd *))
{
	ushort	us;
	int		ret;
#if 1	
	ft_adapter_t	*ft_ada = (ft_adapter_t *) cmd->host->hostdata ;
#endif
	DebScsiFun(printk(KERN_DEBUG "DebScsiFun: ft3xx_queuecommand()\n"););
	cmd->scsi_done = fn ;
	memset(CurScsiReq,0,sizeof(*CurScsiReq));
	CurScsiReq->cmd = (void *)cmd ;
	CurScsiReq->fn  = (void *)fn ;
	CurScsiReq->target = cmd->target + (ft_ada->id * ArrPAda) ;
	CurScsiReq->use_sg = cmd->use_sg ;
	CurScsiReq->buffer = (unchar *)cmd->buffer ;
	CurScsiReq->sense  = (unchar *)cmd->sense_buffer ;
	for( us=0 ; us < 12 ; us ++ )
   		CurScsiReq->cmnd[us] = cmd->cmnd[us] ;
	if(cmd->use_sg) {
		struct scatterlist *sgpnt;
		sgpnt = (struct scatterlist *) cmd->request_buffer;
		for( us=0 ; us < cmd->use_sg ; us++ ) {
			CurScsiReq->WStartAddr[us] = virt_to_bus(sgpnt[us].address) ;
			CurScsiReq->WCount[us]     = sgpnt[us].length ;
		}
		CurScsiReq->WCount[us-1] |= (1<<31);	// EOT=1
	}
	ret = do_queuecommand( CurScsiReq ) ;
	return( ret ) ;
}

/***********************************************************
 * Function:
 *   EndRequest(unchar id,int result)
 *
 * Description:
 *   End SCSI request and return to SCSI high/middle level
***********************************************************/
char EndRequest(void *cmnd,void *fun,int result) 
{
	Scsi_Cmnd	     *cmd  = (Scsi_Cmnd *)cmnd ;
	void	(*fn)(Scsi_Cmnd *) = (void *)fun ;

	cmd->result = result ;
	cmd->scsi_done = fn ;
	fn(cmd);
	DEB(printk(KERN_DEBUG "End SCSI request .\n"));

	return(0);
}

/***********************************************************
 * Function:
 *   ft3xx_abort
 *
 * Description:
 *   cmd - SCSI command.
***********************************************************/
int AbortCommand(void *cmd) ;
int ft3xx_abort(Scsi_Cmnd *cmd)
{
	int	ret ;
	if (cmd->serial_number != cmd->serial_number_at_timeout) {
		return(SCSI_ABORT_NOT_RUNNING);
	}
	ret = AbortCommand(cmd) ;
   	cmd->result = DID_ABORT << 16 ;
   	cmd->scsi_done(cmd);
	if( ret ) {
	   	printk(KERN_ERR "ft3xx: SCSI_ABORT_ERROR\n");
		return(SCSI_ABORT_ERROR);	/* 0x05 */
	} else {
	   	printk(KERN_WARNING "ft3xx: SCSI_ABORT_SUCCESS\n");
		return(SCSI_ABORT_SUCCESS);	/* 0x01 */
	}
}

int ft3xx_abort_eh(Scsi_Cmnd *cmd)
{
	int	ret ;
	if (cmd->serial_number != cmd->serial_number_at_timeout) {
		return(SCSI_ABORT_NOT_RUNNING);
	}
	ret = AbortCommand(cmd) ;
   	cmd->result = DID_ABORT << 16 ;
   	cmd->scsi_done(cmd);
	if( ret ) {
   		printk(KERN_ERR "ft3xx: ft3xx_abort_eh FAILED\n");
		return(FAILED);			/* 0x2003 */
	} else {
   		printk(KERN_WARNING "ft3xx: ft3xx_abort_eh SUCCESS\n");
		return(SUCCESS);		/* 0x2002 */
	}
}

/***********************************************************
 * Function:
 *   ft3xx_reset
 *
 * Description:
 *   cmd - SCSI command.
 *   reset_flags - reset flags.
***********************************************************/
extern int ResetController(void);
int ft3xx_reset(Scsi_Cmnd *cmd, unsigned int reset_flags)
{
	int	ret ;
	ret = ResetController();
   	cmd->result = DID_RESET << 16 ;
   	cmd->scsi_done(cmd);
	if ( ret ) {
		printk(KERN_ERR "ft3xx: SCSI_RESET_ERROR\n");
	       	return(SCSI_RESET_ERROR);
	} else {
		printk(KERN_WARNING "ft3xx: SCSI_RESET_SUCCESS\n");
	       	return(SCSI_RESET_SUCCESS);	/* 0x02 */
	}
}

int ft3xx_reset_eh(Scsi_Cmnd *cmd)
{
	int	ret ;
	ret = ResetController();
   	cmd->result = DID_RESET << 16 ;
   	cmd->scsi_done(cmd);
	if ( ret ) {
		printk(KERN_ERR "ft3xx: ft3xx_reset_eh FAILED\n");
		return(FAILED);		/* 0x2003 */
	} else {
		printk(KERN_WARNING "ft3xx: ft3xx_reset_eh SUCCESS\n");
	       	return(SUCCESS);	/* 0x2002 */
	}
}

/***********************************************************
 * Function:
 *   ft3xx_bios_param
 *
 * Description:
 *   This function returns unit geometry in cylinders/heads/sectors.
 *   input: Geometry[0]/Geometry[1]/Geometry[2] = Heads/Sectors/Cylinders
***********************************************************/
int ft3xx_bios_param(Disk *disk, kdev_t dev, int Geometry[])
{
	Geometry[0]=64;  	
	Geometry[1]=32;  	
	Geometry[2]=disk->capacity / (Geometry[0]*Geometry[1]); 
 
	/* capacity larger than 1GB(C/H/S = 1024/64/32) */
	if (disk->capacity > 0x200000) {
		Geometry[0]=255;
		Geometry[1]=63; 
		Geometry[2]=disk->capacity/(Geometry[0]*Geometry[1]); 
 
		/* capacity larger than 526GB(C/H/S = 65535/255/63) */
		if (Geometry[2] > 65535) {
			Geometry[2] = 65535;
			Geometry[1] = disk->capacity / (255*65535) ;
  		}  
	}
 
   return 0;
}

ulong wV2P(unchar *p_b,unchar *p_wCount)
{
	ulong	addr;

	DebEngFun(printk(KERN_DEBUG "DebEngFun: wV2P(0x%-lx)\n",(ulong)p_b););
	addr = virt_to_bus(p_b);
	return(addr);
}

unchar *p_P2V(ulong wAddr,ushort hCount)
{
	DebEngFun(printk(KERN_DEBUG "DebEngFun: p_P2V(0x%-lx)\n",wAddr););
	return((unchar *)bus_to_virt(wAddr));
}

#ifdef MODULE
static Scsi_Host_Template driver_template = ft3xx;
#include "/usr/src/linux/drivers/scsi/scsi_module.c"
#endif


