/********************************************************************* 
  
   ftlog.c -- PROMISE FastTrak Controllers device driver for Linux.

   Copyright (C) 2002-2005 PROMISE Technology, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   Bugs/Comments/Suggestions should be mailed to:                            
   support@promise.com.tw

   For more information, goto:
   http://www.promise.com

*********************************************************************/

#include <linux/kernel.h>
#include <linux/string.h>
#include "global.h"

void log_snmp_event (int sequence, int eventid, int arrayid, int diskid) 
{
	/*
	 * don't log more than 50 events at a time to prevent running out of
	 * space if ioctl is not called. ioctl resets snmp_log_index
	 * once all events have been sent.
	 */
	if (snmp_log_index < 49) {
    		snmp_log[snmp_log_index].sequence  = sequence;
		snmp_log[snmp_log_index].eventid   = eventid;
		snmp_log[snmp_log_index].arrayid   = arrayid;
		snmp_log[snmp_log_index++].diskid  = diskid;
	}    
}  

void print_event(int eventid, int ArrayID, int dev, unsigned long LBA, 
	int percent, short command, short status) 
{
	static int 	num = 0; 
	char 		diskname[100], str[100];
	int		i;

	num++;
	ArrayID++;
	strcpy(diskname, ft_drive[dev].cam_dev->Model);

	/* cut diskname for showing on console */
	i = 0;
	while( diskname[i] != '\0' ) {
		if( (diskname[i] == ' ') && (diskname[i+1] == ' ') ) {
			diskname[i] = '\0';
			break;
		}
		i++;
	}

	sprintf( str, " (Channel %d/" , ft_drive[dev].pchannel->id+1 );
	strcat(diskname, str );
	strcat(diskname,(ft_drive[dev].select == 0xA0)?"Master)":"Slave)");
  
	//log the event for retreiva thru ioctl
	log_snmp_event(num,eventid, ArrayID, dev+1);

	//	now,latest eventid = 4169
	switch (eventid) {
		
	/*
	 *	Disk Informations:
	 */
		case 0x100: 
			printk(KERN_ERR "ft3xx|%3d|100h|Error|Disk %s marked offline\n", num, diskname);
			break;     
		case 0x101: 
			printk(KERN_ERR "ft3xx|%3d|101h|Information|Disk %s has been attached\n", num, diskname);
			break;     
		case 0x102: 	/* bad sector - but can be repaired */
			printk(KERN_ERR "ft3xx|%3d|102h|Warning|A media error occurred on disk %s\n", num, diskname);
			break;
		case 0x103: 
			printk(KERN_ERR "ft3xx|%3d|103h|Warning|S.M.A.R.T. error reported on disk %s\n", num, diskname);
			break;
		case 0x104: 
			printk(KERN_ERR "ft3xx|%3d|104h|Information|Retry access at disk %s\n", num, diskname);
			break;
		case 0x105: 
			printk(KERN_ERR "ft3xx|%3d|105h|Warning|Disk %s ECC error\n", num, diskname);
			break;
		case 0x106: 
			printk(KERN_ERR "ft3xx|%3d|106h|Warning|Disk %s time out\n", num, diskname);
			break;
#if 0
		case 0x107: 	
			printk(KERN_ERR "ft3xx|%3d|107h|Error|Disk %s error at LBA 0x%lx starting from LBA 0x%lx (Feature=0x%lx,Block Count=0x%lx,Command=0x%lx,Status=0x%lx,Error=0x%lx)\n", num, diskname, .....);
			break;     
		case 0x108: 
			printk(KERN_ERR "ft3xx|%3d|108h|Information|Bad Sector log updated: LBA 0x%lx on Disk %s\n", num, LBA, diskname);
			break;
#endif
		case 0x109: 
			printk(KERN_ERR "ft3xx|%3d|109h|Information|Channel %d was reset\n", num, (dev/2)+1);
			break;
#if 0
		case 0x10A: 
			printk(KERN_ERR "ft3xx|%3d|10Ah|Information|Bad sector log entry cleared: LBA 0x%lx on disk %s\n", num, LBA, diskname);
			break;
		case 0x10B: 
			printk(KERN_ERR "ft3xx|%3d|10Bh|Error|Reserved Sector(Metadata) Error: LBA 0x%lx on disk %s\n", num, LBA, diskname);
			break;
		case 0x10C: 
			printk(KERN_ERR "ft3xx|%3d|10Ch|Error|Bad Sector on Array %s: LBA 0x%lx on disk %s\n", num, ArrayID, LBA, diskname);
			break;
#endif
	
	/*
	 *	Array Informations:
	 */
		case 0x200: 	/* bad sector - can't be repaired */
			printk(KERN_ERR "ft3xx|%3d|200h|Error|Disk %s in Array %d has read/write error (LBA=0x%lx Command=0x%x Status=0x%x)\n", num, diskname, ArrayID, LBA, command, status);
			break;
		case 0x201: 
			printk(KERN_ERR "ft3xx|%3d|4132|Error|Array %d offline\n", num, ArrayID);
			break;     
		case 0x202: 
			printk(KERN_ERR "ft3xx|%3d|202h|Warning|Array %d Critical\n", num, ArrayID);
			break;     
		case 0x203: 
			printk(KERN_ERR "ft3xx|%3d|203h|Information|Create new array %d\n", num, ArrayID);
			break;     
		case 0x204: 
			printk(KERN_ERR "ft3xx|%3d|204h|Warning|Delete array %d\n", num, ArrayID);
			break;     
		
	/*	
	 *	Rebuild Information: 
	 *	Start/AutoStart/Pause/Stop/Abort/Resume/Complete 
	 */
		case 0x210: 
			printk(KERN_ERR "ft3xx|%3d|210h|Information|Rebuilding started on disk %s of Array %d\n", num, diskname ,ArrayID);
			break;
		case 0x211: 
			printk(KERN_ERR "ft3xx|%3d|211h|Information|Automatic rebuilding started on disk %s of Array %d\n", num, diskname ,ArrayID);
			break;
		case 0x212: 
			printk(KERN_ERR "ft3xx|%3d|212h|Warning|Rebuilding aborted on Array %d at %d percent completed\n",num,ArrayID,percent);
			break;
		case 0x213: 
			printk(KERN_ERR "ft3xx|%3d|213h|Warning|Rebuilding paused on Array %d at %d percent completed\n",num,ArrayID,percent);
			break;
		case 0x214: 
			printk(KERN_ERR "ft3xx|%3d|214h|Information|Rebuilding resumed on Array %d at %d percent completed\n",num,ArrayID,percent);
			break;     
		case 0x215: 
			printk(KERN_ERR "ft3xx|%3d|215h|Information|Rebuilding successfully completed for disk %s in Array %d\n", num, diskname, ArrayID);
			break;
		case 0x216: 
			printk(KERN_ERR "ft3xx|%3d|216h|Error|Rebuilding failed on Array %d at %d percent completed\n",num,ArrayID,percent);
			break;

	/*	
	 *	Synchronize Information: 
	 *	Start/AutoStart/Pause/Stop/Abort/Resume/Complete 
	 */
		case 0x220: 
			printk(KERN_ERR "ft3xx|%3d|220h|Information|Synchronization started on Array %d\n", num, ArrayID);
			break;     
		case 0x221: 
			printk(KERN_ERR "ft3xx|%3d|221h|Information|Synchronization aborted on Array %d at %d percent completed\n",num,ArrayID,percent);
			break;
		case 0x222: 
			printk(KERN_ERR "ft3xx|%3d|222h|Information|Synchronization paused on Array %d at %d percent completed\n",num,ArrayID,percent);
			break;     
		case 0x223: 
			printk(KERN_ERR "ft3xx|%3d|223h|Information|Synchronization resumed on Array %d at %d percent completed\n",num,ArrayID,percent);
			break;     
		case 0x224: 
			printk(KERN_ERR "ft3xx|%3d|224h|Information|Synchronization successfully completed on Array %d\n", num, ArrayID);
			break;     
		case 0x225: 
			printk(KERN_ERR "ft3xx|%3d|225h|Error|Synchronization failed on Array %d at %d percent completed\n",num,ArrayID,percent);
			break;
		case 0x226: 
			printk(KERN_ERR "ft3xx|%3d|226h|Error|Synchronization comparison error on Array %d at %d percent completed\n",num,ArrayID,percent);
			break;
			
	/*	
	 *	Conversion/Expansion Information: 
	 *	Start/AutoStart/Pause/Stop/Abort/Resume/Complete 
	 */
		case 0x230: 
			printk(KERN_ERR "ft3xx|%3d|230h|Information|Expansion/Conversion begun on Array %d\n", num, ArrayID);
			break;     
		case 0x231: 
			printk(KERN_ERR "ft3xx|%3d|231h|Information|Expansion/Conversion continued on Array %d\n", num, ArrayID);
			break;     
		case 0x232: 
			printk(KERN_ERR "ft3xx|%3d|232h|Information|Expansion/Conversion successfully completed on Array %d\n", num, ArrayID);
			break;     
		case 0x233: 
			printk(KERN_ERR "ft3xx|%3d|233h|Warning|Expansion/Conversion aborted on Array %d at %d percent completed\n", num, ArrayID,percent);
			break;     
		case 0x234: 
			printk(KERN_ERR "ft3xx|%3d|234h|Warning|Expansion/Conversion paused on Array %d at %d percent completed\n", num, ArrayID,percent);
			break;     
		case 0x235: 
			printk(KERN_ERR "ft3xx|%3d|235h|Information|Expansion/Conversion resumed on Array %d at %d percent completed\n", num, ArrayID,percent);
			break;     
		case 0x236: 
			printk(KERN_ERR "ft3xx|%3d|236h|Error|Expansion/Conversion failed on Array %d at %d percent completed\n", num, ArrayID,percent);
			break;     
		case 0x237: 
			printk(KERN_ERR "ft3xx|%3d|237h|Information|Array %d initialization started\n", num, ArrayID);
			break;     
		case 0x238: 
			printk(KERN_ERR "ft3xx|%3d|238h|Information|Synchronization started on uninitialized Array %d\n", num, ArrayID);
			break;     

	/*
	 *	Swap Box Event
	 *	Temperature/Fan RPM/Voltage(5/12) 
	 */
		case 0x300: 
			printk(KERN_ERR "ft3xx|%3d|300h|Warning|Temperature is too high for enclosure %d. The temperature is %ld C/ %ld F\n", num, ArrayID, LBA, ((9*LBA)/5)+32 );
			break;     
		case 0x301: 
			printk(KERN_ERR "ft3xx|%3d|301h|Warning|Fan stopped for enclosure %d.\n", num, ArrayID);
			break;     
		case 0x302: 
			printk(KERN_ERR "ft3xx|%3d|302h|Warning|5V over/under threshold for enclosure %d. The voltage is %ldmV\n", num, ArrayID, LBA);
			break;     
		case 0x303: 
			printk(KERN_ERR "ft3xx|%3d|303h|Warning|12V over/under threshold for enclosure %d. The voltage is %ldmV\n", num, ArrayID, LBA);
			break;     
		case 0x304: 
			printk(KERN_ERR "ft3xx|%3d|304h|Warning|3.3V over/under threshold for enclosure %d. The voltage is %ldmV\n", num, ArrayID, LBA);
			break;     
		case 0x305: 
			printk(KERN_ERR "ft3xx|%3d|305h|Information|Enclosure %d power on\n", num, ArrayID);
			break;     
		case 0x306: 
			printk(KERN_ERR "ft3xx|%3d|306h|Information|Enclosure %d power off\n", num, ArrayID);
			break;     
		default:
			printk(KERN_ERR "ft3xx: print_event(%#x) IE\n",eventid);
	}
}

void log_event(int eventid, int ArrayID, int dev, unsigned long LBA, 
	int percent, short command, short status) 
{
	/* print monitoring kernel messages to console */
	print_event(eventid, ArrayID, dev, LBA, percent, command, status) ;

	/* 
	 *  Issue-Signal timer to notify RAID-API, and then 
	 *  RAID-API will use IOCTL to get global data "gLogEvent[]" 
	 */
 	AddIssueSignalTimer(eventid);

	return;
}
      
