/********************************************************************* 
  
   global.h -- PROMISE FastTrak Controllers device driver for Linux.

   Copyright (C) 2002-2005 PROMISE Technology, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   Bugs/Comments/Suggestions should be mailed to:                            
   support@promise.com.tw

   For more information, goto:
   http://www.promise.com

*********************************************************************/

#include <linux/blk.h>	
#include <linux/pci.h>	
#include "cam/cam_con.h"
#include "cam/cam_def.h"
#include "cam/cam_exp.h"

#define VERSION	"1.00.0.19"
/* #define VERDATE "(09.JUL.2003)" */
#define VERDATE ""

#define ArrPAda		4	
#define ChPAda		4	
#define DevPCh		2	

#define FT_Max_Ada	2
#define FT_Max_Arr	( FT_Max_Ada * ArrPAda )	/* 8 */
#define FT_Max_Ch	( FT_Max_Ada * ChPAda )		/* 8 */
#define FT_Max_Dev	( FT_Max_Ch  * DevPCh )		/* 16 */
#define	Can_Queue	12
#define Max_SG_Table	32	/* Engine : max bSG_Count = 255 */

/* Define PDC20xxx Device ID */
#define DID319		0x3319	/* FastTrak S150 TX4 */
#define DID371		0x3371	/* FastTrak S150 TX2 */
#define DID376		0x3376	/* FastTrak 376 */
#define DID378		0x3373	/* FastTrak 378 */
#define DID619		0x6629	/* FastTrak TX4000 */
#define DIDEND		0
extern unsigned short 	FT_DEVID[6] ;

#define DEB(x) 			/*	Display messages */
#define	DEBFUN(x) 	 	/*	DEBUG Function	*/
#define	DebFtFun(x)  		/*	DEBUG FastTrak Function	*/
#define	DebEngFun(x) 		/*	DEBUG Engine   Function	*/
#define	DebScsiFun(x) 		/*	DEBUG SCSI     Function	*/

typedef struct ft_adapter ft_adapter_t ;
typedef struct ft_channel ft_channel_t ;
typedef struct ft_drive   ft_drive_t ;

/* Bits of HD_STATUS */
#define ERR	0x01
#define IDX	0x02
#define CORR	0x04
#define DRQ	0x08
#define SEEK	0x10
#define DF	0x20
#define DRDY	0x40
#define BSY	0x80

#define READ_PIO	0x20
#define WRITE_PIO	0x30
#define READ_DMA	0xC8
#define WRITE_DMA	0xCA
#define READ_DMA_EXT	0x25	
#define WRITE_DMA_EXT	0x35	

struct ft_adapter {
	unsigned char	id; 
	unsigned char	present; 
	unsigned short	deviceid;
	unsigned short	irq;
	struct pci_dev	*ft_pci_dev;
	unsigned int	base[6];
	unsigned int	range[6];	/* base address range */
	unsigned int	rom_base;
	unsigned long	cam_mem_addr;	/* virtual mem address for CAM */
	int 		cam_mem_power;	/* __get_free_pages()  for CAM */
//	spinlock_t	ft_lock;
	ft_channel_t	*pchannel[ChPAda];	
	ADAPTER_CONFIG_INFO	*cam_ada;	/* cam adapter info */
}  ;

struct ft_channel {
	ft_adapter_t 	*padapter ;	
	unsigned char	id;
	unsigned char	abnormality;	
	unsigned short	base;
	unsigned short	ctrl;
	unsigned short	bm;
	ft_drive_t	*pdrive[DevPCh];   	
	CHNL_CONFIG	*cam_ch;	/* cam channel info */
}  ;

/* The Main disk structure for each drive */
struct ft_drive {
	ft_channel_t	*pchannel;	
	unsigned char	id;
	unsigned char	select;
	unsigned char	present;
	unsigned char 	downmode;	
	unsigned char 	xfermode;
	unsigned char	arr;
	unsigned char	strm;
	unsigned char	retried;
	unsigned long	lbusy;	
	unsigned long	rebuilding;	
	unsigned long	synchronizing;
	unsigned long	con_exp;
	unsigned char	LBA_48bit;
	unsigned char	box_present;	
	struct ft_driveid	*identify;
	DEV_CONFIG		*cam_dev;	/* cam device info */
}  ;

typedef struct {
	unsigned short	deviceid;
	unsigned char	irq;
	unsigned short	base[5];
	unsigned int 	rom;
}  ft_ladainfo_t ;

typedef struct {
	unsigned char	ID;
	unsigned char	Flag;
	unsigned char 	RAID;
	unsigned char	StripeBlockPower;
	unsigned short	Cyl;
	unsigned char	Hd;
	unsigned char	Sctr;
	unsigned int	Size;
	unsigned int	Watermark;
	unsigned char	strm;
}  ft_larrinfo_t ;

typedef struct {
	unsigned char	ver[30];
	ft_ladainfo_t	adainfo[FT_Max_Ada];
	ft_larrinfo_t	arrinfo[FT_Max_Arr];
	unsigned char	rebdev;
}  ft_info_t ;

struct	scsiengreq {	// SCSI - Engine request
	void		*cmd;
	void		*fn;
	unsigned char	target;
	unsigned char	use_sg;
	unsigned char	*buffer;
	unsigned char	*sense;
	unsigned char	cmnd[12];
	unsigned long	WStartAddr[Max_SG_Table];
	unsigned long	WCount[Max_SG_Table];   
} ; 
typedef	struct	scsiengreq *ScsiEngReq_p ;

typedef struct  {
	char *buffer;
	int  length;
	int  offset;
	int  position;
} ft_proc_t ;

/* atacmd */
extern int GetSmartStatus(int bID, unsigned char *status);

/* global */
extern 	ft_adapter_t 	ft_adapter[FT_Max_Ada];
extern 	ft_channel_t 	ft_channel[FT_Max_Ch];
extern 	ft_drive_t 	ft_drive[FT_Max_Dev];
extern	ft_info_t 	ft_info;
extern	CAM_INFO	ft_cam_info[FT_Max_Ada];

#define bPOWER_UP		0x00
#define bHARD_RESET		0x01
#define bSOFT_RESET		0x02

int 		SleepWakeUp(int id, int fun) ;
unsigned char  	bInitRAID(unsigned char bType,unsigned long wP, unsigned char *p_b, unsigned long wInitInAcceMemSize, unsigned long wInitAcceMemSize,unsigned char AdapterID);

/* fasttrak */
unsigned char	*p_P2V(unsigned long wAddr, unsigned short hCount);

/* ftlog */
void 	log_event(int eventid, int ArrayID, int dev, unsigned long LBA, int percent, short command, short status);

/* os */
extern unsigned char	SCSIRequestFlag ;
extern unsigned long	StartRebJiffies ;
void 			InitQueueBuffer(void) ;

/* utility */
#define	MaxLogEventCount	20
char 	InitGDS( void ) ;
char 	DefDriveStruct (void) ;
int	SetArrayLED(int ArrayID) ;
int 	SetEnclosureLED(int DiskID, int color) ;
int 	GetEngNeededMemSize(ulong *phy,ulong *vir) ;
int	OpenCAM(void) ;
int	CloseCAM(void) ;
int 	StopRebuild(void) ;
unchar 	ChNumPerApt(void) ;
void 	*GetFreeLogQueue(void);
int 	ReleaseUsedLogQueue(void);

/* proc */
int 	GetProcInfo( ft_proc_t *buffer ) ;

/* swap */
int 	FixedCAMDetectEnclosure(void);
int 	CheckEnclosureStatus(void);

/* timer */
int 	bInitRAIDTimer(int AdapterID) ;
int 	RegisterTimer(void) ;
int 	ReleaseTimer(void) ;
int 	AddIssueSignalTimer(int EventID) ;
extern  ulong	ft_seconds ;

/* event log struct provided thru ioctl */
typedef struct {
	int sequence;
	int eventid;
	int arrayid;
	int diskid;
} snmp_log_t;

extern  snmp_log_t	snmp_log[50];
extern  int		snmp_log_index;

typedef struct {
	int 	eventid;
	struct 	fasync_struct *async_queue;
} ftdev_fd_t;
extern	ftdev_fd_t	ftdev_fd;


