/********************************************************************* 
  
   wrapper.c -- PROMISE FastTrak Controllers device driver for Linux.

   Copyright (C) 2002-2005 PROMISE Technology, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   Bugs/Comments/Suggestions should be mailed to:                            
   support@promise.com.tw

   For more information, goto:
   http://www.promise.com

*********************************************************************/

#include <asm/io.h>
#include <asm/ptrace.h>
#include <asm/delay.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/blk.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include "global.h"

/* used same as list_head in linux kernel */
struct 	ft_list_head {
	struct ft_list_head *next, *prev;
};

/* used same as timer_list in linux kernel */
struct 	ft_timer_list {
	struct ft_list_head list;
	unsigned long expires;
	unsigned long data;
	void (*function)(unsigned long);
};

extern void ft_handler(int irq, void *dev_id);

unchar ft_inb (ushort port) {
  	return inb(port);
}

unsigned ft_inw (ushort port) {
	return inw(port);
}

void ft_outb (unchar value, ushort port) {
  	outb(value, port);
}

void ft_outw (ushort value, ushort port) {
	outw(value, port);
}

void ft_outl (ulong value, ushort port) {
  	outl(value, port);
}

/* int ft_printk(const char *fmt, ...) { */
int ft_printk(const char *fmt, va_list *argp) {
 	return printk(fmt, argp);
}

/* extern int sprintf(char * buf, const char * fmt, ...); */
int ft_sprintf(char * buf, const char * fmt, va_list *argp) {
	return sprintf(buf, fmt, argp);
}

void ft_init_timer(struct ft_timer_list * timer) {
/* If necessary convert ft_timer_list into timer_list */
  	init_timer((struct timer_list *)timer);
}

void ft_add_timer(struct ft_timer_list * timer) {
/* If necessary convert ft_timer_list into timer_list */
  	add_timer((struct timer_list *)timer);
}

void ft_del_timer(struct ft_timer_list * timer) {
/* If necessary convert ft_timer_list into timer_list */
  	del_timer((struct timer_list *)timer);
}
  
void ft_udelay(unsigned long usecs) {
  	udelay(usecs);
}

void ft_mdelay(unsigned long msecs) {
  	mdelay(msecs);
}

unsigned long volatile ft_get_jiffies(void) {
/* jiffies is the number of clock ticks since the OS was booted */
/* it is exported by the linux kernel */
  	return jiffies;
}

void ft_do_gettimeofday(struct timeval *tv) {
	do_gettimeofday(tv);
}

int ft_get_HZ(void) {
	return(HZ);
}

void *ft_ioremap( unsigned long phys_address , unsigned long size){
	return ioremap( phys_address, size ); 
}

void ft_iounmap( void *mem_address ){
	iounmap( mem_address );
	return;
}

int ft_pci_read_config_word( struct pci_dev *dev, int where, u16 *val ){
	return pci_read_config_word( dev, where, val );
}

int ft_pci_write_config_word( struct pci_dev *dev, int where, u16 val ){
	return pci_write_config_word( dev, where, val );
}

int ft_vsprintf( char *buf, const char *str , va_list list){
	return vsprintf( buf, str, list );
}

void *ft_kmalloc(unsigned int size,int priority) {
	return(kmalloc(size,priority));
}

void ft_kfree(void *obj) {
	kfree(obj);
}

void ft_free_irq(int irq, void *obj ) {
	free_irq(irq, obj) ;
}

void ft_panic(const char *fmt, va_list *argp ) {
 	panic(fmt, argp);
}

void ft_release_region(unsigned int base, unsigned int range) {
	release_region(base, range);
}

void ft_free_pages(unsigned long addr, int order) {
	free_pages(addr, order);
}

unsigned long ft__get_free_pages(unsigned int gfp_mask, unsigned int order ) {
	return (__get_free_pages(gfp_mask, order));
}

void ft_kill_fasync(struct fasync_struct **async_qp, int sigio, int pollin) {
	kill_fasync(async_qp, sigio, pollin);
}
